( Sounds Program for 3-instruction Forth on Atmel 328 Arduino Nano )
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2020-06-22, 2020-07-25 edited
\ 2020-10-03---Sat---22:00 - Expanded to include the lights capability with Piano-4x4

( * WORDS FOR COMPUTERIZED MUSIC * )

\ Hardware Options:
\   Buzzer: 5V,  + on pin, - on ground, use tone 440 --- loud sound
\   8 OHM Speaker with 250R or greater resistor.  resistor between pin and speaker+, speaker - to ground --- small sound
\   8 OHM Speaker with 2N2222 transistor for amplification.  5V to collector, pin to base, emitter to spkr+, spkr- to GRD-- big sound


\ Design: music is a time series of NOTES (frequencies), their DURATIONS (ms), and the RESTS (ms) between them.

\ Simplest play (v1) is specifying the frequencies and using fixed duration (120ms) and fixed rest (120-240ms).

\ Better (v2) is three arrays holding notes frequencies, note durations, and rests.  Play interleaves the instrument with all three.

\ AE - 2021-05-12---Wed---05:34  --- REMOVING the need for 3iLight.fs to just use this for the sound capability
: _light_ ;
: light| ;
: out4 ;
: use-port-out ;


variable SPEAKERpin  \ note - this is speaker for the PROGRAMMABLE SYNTHESIZER.  AVR 3iForth sets a DIGITAL PIANO speaker.  
                     \ If the two are different, then can have two speakers and voices...

: setspeaker ( pin -- ) SPEAKERpin ! ;
: speaker ( -- pin ) SPEAKERpin @ ;

variable dur_ms 120 dur_ms ! \ basic beat duration ms  (sixteenth)
: beat dur_ms @ ; 
: setbeat ( n -- ) dur_ms ! ;
    
: R1 beat ms ;
: R2 beat 2 * ms ; \ rest

( * MUSICAL NOTES   --   provides frequency and generates tone * )
\ Calculation for the half steps:  
\ The mathematics of western music is multiplicative (equitempered scale popularized by Bach in the 1700s)
\ The musical scale has 12 semitones with each octave being double the frequency of the previous
\ So the ratio between successive musical semitones is fixed as 12th-root of 2 = 2^(1/12) = 1.05946309436
\ By convention, there are 2 flats (b%, e%) and three sharps (c#, f#, g#)
\ 73 semitones covering 6 octaves (=12*6+1)
create piano   \ 108 semitones, spanning 9 octaves, each a dword (4 bytes)
    \       C     d%     D      e%     E      F      F#     G      a%     A      b%     B
    \             1             3                    6             8             10
    \       0            2             4      5             7             9            11      
\ ( 0 )      16 ,   17 ,   18 ,   19 ,   21 ,   22 ,   23 ,   25 ,   26 ,   28 ,   29 ,   31 , 
\ For Arduino Atmel 328 - It is not possible to generate tones lower than 31Hz (B0)
( 0 )      33 ,   35 ,   37 ,   39 ,   41 ,   44 ,   46 ,   49 ,   52 ,   55 ,   58 ,   62 ,
( 1 )      65 ,   69 ,   73 ,   78 ,   82 ,   87 ,   93 ,   98 ,  104 ,  110 ,  117 ,  124 ,
( 2 )     131 ,  139 ,  147 ,  156 ,  165 ,  175 ,  185 ,  196 ,  208 ,  220 ,  233 ,  247 ,
( 3 )     262 ,  277 ,  294 ,  311 ,  330 ,  349 ,  370 ,  392 ,  415 ,  440 ,  466 ,  494 ,  \ middle
( 4 )     523 ,  554 ,  587 ,  622 ,  659 ,  698 ,  740 ,  784 ,  831 ,  880 ,  932 ,  988 , 
( 5 )    1047 , 1109 , 1175 , 1245 , 1319 , 1397 , 1480 , 1568 , 1661 , 1760 , 1865 , 1976 , 
( 6 )    2093 , 2217 , 2349 , 2489 , 2637 , 2794 , 2960 , 3126 , 3322 , 3520 , 3729 , 3951 , 
( 7 )    4186 , 4435 , 4699 , 4978 , 5274 , 5588 , 5920 , 6272 , 6645 , 7040 , 7459 , 7908 ,
96 constant semitones  \ spanning 8 octaves

: f2i  ( freq -- idx )  \ find index given frequency
   >r piano begin @+ r@ = until r> drop  \ -- pointer
   piano -  4 bytes /   1 -  ;

create piano-led-map  \ maps led codes to led numbers, 0-based, to match switch/button numbering
    $c c, $b c, $a c, $9 c, $8 c, $7 c, $6 c, $5 c, 
    $4 c, $3 c, $2 c, $1 c, $0 c, $f c, $e c, $d c, 
: piano-test  \ tests each light, prompted by key presses
    piano-led-map 16 0 do c@+ dup . _light_ key drop loop drop light| ;
: piano-flash
    piano-led-map 16 0 do c@+ _light_ loop drop light| 50 ms ;
: piano-flash-n
    0 do piano-flash loop ;

: __ ( led# -- ) \ turn on appropriate light, 0-based
    piano-led-map + c@ _light_ ;
: f2light ( freq -- light ) 
    f2i 12 mod __ ;

create scale-led-map \ maps 1-8 to the appropriate lights, to match major key numbers
    0 c, 2 c, 4 c, 5 c, 7 c, 9 c, 11 c, 12 c, \ white keys
    1 c, 3 c, 6 c, 8 c, 10 c,                 \ black keys
: #_ ( scale# -- led# ) \ which led mapped code to turn on for scale based input
    1 - scale-led-map + c@ ;
: #__ ( scale# -- ) #_ __ ; \ turn on appropriate light, scale based
   
\ Sharps and Flats based on the 8-key numbering, with sharps and flats, indexed into the key-map above.
: 1# 9  ;
: 2% 9  ;
: 2# 10 ;
: 3% 10 ;
: 4# 11 ;
: 5% 11 ;
: 5# 12 ;
: 6% 12 ;
: 6# 13 ;
: 7% 13 ;

: tone ( freq -- ) ( dup f2light ) lohi beat lohi speaker [tone] R2 ;  \ CODE REVIEW: should lights play within tone?
    \ 13 setspeaker
    \ 440 tone
: ring ( -- ) 440 lohi 500 lohi speaker [tone] ; 
: ring-ring  ( -- )  ring 600 ms ring ; \ ring tone

\ CODE REVIEW: IS THIS REALLY NEEDED??
\ Playing by the index into piano array  ( 0 to 72 )
: freq ( idx -- )  dwords piano + @ ; \ ex: 12 freq tone 
\ Playing by 2-D coordinations
: 2di ( irow jcol -- idx )  swap 12 * + ;  \ 2d coordinates to 1d index, e.g. 2 10 2di freq tone
        
\ CODE REVIEW: IS THIS REALLY NEEDED??
\ Playing by root and offsets (relative numbers)
variable zero 
: set-zero ( i j -- ) 2di zero c! ;  \ ex: 3 0 set-zero
: reset-zero 0 0 set-zero ;
reset-zero
: show-zero  ( -- idx ) zero c@ ;

\ CODE REVIEW: IS THIS REALLY NEEDED??
\ Playing music 
: note ( idx -- ) show-zero + freq tone ;
: note2d ( irow icol -- ) 2di note ;
\ Playing music by number relative to note-offset (show-zero)
: ., ( num1-8 -- ) 1 - note ;

\ TONES DEMO 
: upward  semitones 0 do i note loop ;
: downward semitones 0 do semitones 1 - i - note loop ;
: updown upward downward ;
: updown-fast 60 setbeat updown 120 setbeat ;
: updown-sfast 30 setbeat updown 120 setbeat ;
: updown-slide 15 setbeat updown 120 setbeat ;


\ Playing by reference against a scale
variable scale-ptr    \ holds ptr to scale, allowing playing tunes against 8-key predefined scale
: scale  ( -- a* )  scale-ptr @ ;
: set-scale ( a* -- )  scale-ptr ! ;

\ these are indices into the piano table -- e.g. 36 note    
create cstd8 36 c, 38 c, 40 c, 41 c, 43 c, 45 c, 47 c, 48 c,   \ white keys
    37 c, 39 c, 42 c, 44 c, 46 c,  \ black keys    equitemper
create cstd5 37 c, 39 c, 42 c, 44 c, 46 c, 
create cmaj8 31 c, 33 c, 36 c, 38 c, 40 c, 43 c, 45 c, 48 c, 
create dmin8  38 c, 40 c, 41 c, 43 c, 45 c, 46 c, 49 c, 50 c,  ( d. e. f. g. a. b%. c#' d' )
create bhairav8  31 c, 32 c, 35 c, 36 c, 38 c, 39 c, 42 c, 43 c,

: checkfreq checkreg rcv-buffer c@+ swap c@ swap 256 * + ; \ use for checking Piano frequencies

create black-registers 1 c, 3 c, 6 c, 8 c, 10 c, 
create white-registers 0 c, 2 c, 4 c, 5 c, 7 c, 9 c, 11 c, 12 c,

: black-null  ( -- ) 5 0 do 0 black-registers i + c@ setreg loop ;
: xscale5 ( scale -- )  \ ex: cstd5 xscale5
    5 0 do dup i + c@ freq black-registers i + c@ setreg loop drop ;
: xscale8 ( scale -- )  \ send scale to target  ex: cmaj8 xscale8
    8 0 do dup 
        i + c@ freq
        white-registers i + c@ 
        setreg 
    loop black-null drop ;

: cstd  cstd8 set-scale  cstd8 xscale8  cstd5 xscale5 ;
: cmaj  cmaj8 set-scale  cmaj8 xscale8 ;
: bhairav bhairav8 set-scale bhairav8 xscale8 ;
: dmin  dmin8 set-scale  dmin8 xscale8 ;
    
: .. ( n -- ) \ plays note using scale array, 1 indexed
    1 - scale + c@ note ;
: play-scale  9 1 do i .. loop ;



: start-speaker ( programmable digital synthesizer )
    13 out
    13 setspeaker
    reset-zero
	cstd play-scale  ;    \ nice chord: b.. d. g. 

: start-piano4x4  ( programmable digital synthesizer with lights board )
    start-speaker
    out4 use-port-out  ;


\ ETHNIC SYNTHESIZERS

\ setting up ethno tunes
: whites ( 1 2 3 4 5 6 7 8 -- )  \ setup white frequency-key mappings on target piano
    12 setreg 11 setreg 9 setreg 7 setreg 5 setreg 4 setreg 2 setreg 0 setreg ;
: blacks ( 1 2 3 4 5 -- )
    10 setreg 8 setreg 6 setreg 3 setreg 1 setreg ;
: null-black-keys ( -- )
    0 0 0 0 0 blacks ;

: init-keys
   10 13 setreg  \ set min duration (ms) of note
   5 14 setreg  \ set settle time following note
;
: equitemper  \ Bach's equitemper scale, do-re-me tuning
   262  294   330   349   392   440   494   523   whites
   277  311   370   415   466   blacks  ;
   
: set-cmaj  \ starts on G  She'll be coming round the mountain when she comes...   O! Susanna (start on 3)
    196 220 262 294 330 392 440 523
    whites null-black-keys ;

\ Music theory:
\ Key signature: https://en.wikipedia.org/wiki/Key_signature
\   There are fifteen homogeneous key signatures ranging from 7 flats to 7 naturals to 7 sharps, with each successive flat or sharp placed on the note a perfect fifth below or above the previous one
\ Keys Major and Minor: https://en.wikipedia.org/wiki/Major_and_minor#Keys
\ In Western music, there are 12 semitones (half steps).  A scale picks out 8 of these.
\ A major scale has a HALFSTEP between notes 3 & 4 and between notes 7 & 8  (eg C major between E-F and B-C)
\ A minor scale puts the HALFSTEP between notes 2 & 3 and between notes 5 & 6 (eg D minor between E-F and A-B%)
\ Octave is a 2:1 ratio, Fifth is a 3:2 ratio
\ Intonation: 12-tone equal temperament is the most common tuning system in the West

\ INDIAN SCALES
\ Source: Special Topics in Music Theory, by Catherine Schmidt-Jones
\    Ch. 3.5 Indian Classical Music: Tuning and Ragas
\ https://cnx.org/contents/mviDomhA@5.66:bDLyHWWs@16/Indian-Classical-Music-Tuning-and-Ragas
\ A "that" is a scale of 7 notes -- why 7?  (7 celestial bodies - Sun, Moon, Mercury, Venus, Mars, Jupiter and Saturn)
\ Aside: 7 days a week: Babylonians of the 6th century... https://www.discovermagazine.com/planet-earth/why-are-there-seven-days-in-a-week
\ A raga chooses five, six, or all 7 notes of the that, and sticks to that, i.e. pentatonic raga, hexatonic raga, or full raga

: bhairav.  \ raga with two flats on D and A    Sa Re Ga Ma Pa Dha Ni   NO EQUIVALENT WESTERN MAJOR SCALE
   262   277   330   349   392   415   494  523   whites  null-black-keys ;
: asavari  \ raga with three flats on E, A, B    Sa Re Ga Ma Pa Dha Ni   WESTERN: EB-MAJOR  ex: stairway to heaven
   262   294   311   349   392   415   466   523  whites  null-black-keys ;
: bhairavi  \ raga with four flats on D, E, A, B    Sa Re Ga Ma Pa Dha Ni    WESTERN: AB-MAJOR
   262   277   311   349   392   415   466   523 whites  null-black-keys ;
: kafi  \ raga with two flats on E and B    Sa Re Ga Ma Pa Dha Ni   ex: unchaare kot 3rd note, aaj aanand 1st note, and scarborough fair   WESTERN: BB-MAJOR
   262   294   311   349   392   440   466   523  whites  null-black-keys ;
: khamaj \ raga with 1 flat on B   WESTERN: F MAJOR
   262   294   330   349   392   440   466   523  whites  null-black-keys ;
: kalyan \ raga with 1 sharp on F   WESTERN: G MAJOR
   262    294    330    370    392    440    494    523  whites  null-black-keys ;
: purvi \ raga with 1 sharp on F and two flats on D and A  "back to life, back to reality"
   262    277     330    370    392    415    494    523  whites  null-black-keys ;
: todi  \ bhairavi raga (four flats on D, E, A, B) with additionally 1 sharp on F    Sa Re Ga Ma Pa Dha Ni
   262    277    311    370    392    415    466    523  whites  null-black-keys ;
: bilawal. ; \ raga with no flats or sharps, ex: abadoo and doh-a-deer  \ Western C-major 

\  CHINESE SCALES

: yu  \ Yu scale on Huang Zhong base freq ( Why Don't you come back... and Fire on the Mountain)
   440    529    594    660    793    440 2 *   529 2 *   594 2 *  whites  null-black-keys ;
: yu-1  \ Yu scale on Huang Zhong base freq shifted down one octave ( Fire on the Mountain )
   793 1 rshift   440   529  594   660   793  440 2 *   529 2 *  whites  null-black-keys  ;
: shang  \ Shang scale on Da Cu  base freq ( Snoopy theme)
   495    557    669    743    892    495 2 *     557 2 *    669 2 * whites  null-black-keys ;
: gong  \ Gong scale on Zhong Lu base freq
   594    669    752    892    1003    594 2 *   669 2 *   752 2 *   whites null-black-keys ;
: jue  \ Jue scale on Yi Ze base freq 
   705 0    847 2    952 4    1129 5   1270 7   705 2 *    847 2 *    952 2 * ;
: zhi  \ Zhi scale on Wu Yi base freq 
   793    892    1071   1189   1338   793 2 *    892 2 *    1071 2 *  ;



\ Frequency lookup table - basic
\ Note double is an octave higher, and half is an octave lower -- CONSIDER using fixed point and calculating the initial table
\ Flats and sharps as + / - against base
: C...    65 tone ;  \ octave lower 2
: c#...   69 tone ;
: d%...   69 tone ; 
: D...    73 tone ;
: d#...   78 tone ;
: e%...   78 tone ;
: E...    82 tone ;
: F...    87 tone ;
: f#...   93 tone ;
: g%...   93 tone ;
: G...    98 tone ;     
: g#...  104 tone ;
: a%...  104 tone ;
: A...   110 tone ;
: a#...  117 tone ;
: b%...  117 tone ;
: B...   124 tone ;
: C..    131 tone ;  \ octave lower
: c#..   139 tone ;
: d%..   139 tone ;
: D..    147 tone ;
: d#..   156 tone ;
: e%..   156 tone ;
: E..    165 tone ;
: F..    175 tone ;
: f#..   185 tone ;
: g%..   185 tone ;
: G..    196 tone ;     
: g#..   208 tone ;
: a%..   208 tone ; 
: A..    220 tone ;       
: a#..   233 tone ;
: b%..   233 tone ;
: B..    247 tone ;
: C.     262 tone ; \ middle octave -- Mddle C
: c#.    277 tone ;
: d%.    277 tone ;
: D.     294 tone ; 
: d#.    311 tone ;
: e%.    311 tone ;
: E.     330 tone ;        \ tone(SPEAKER,330,QUARTER); delay(SHORT_DELAY); } 
: F.     349 tone ; 
: f#.    370 tone ;
: g%.    370 tone ; 
: G.     392 tone ;
: g#.    415 tone ;    
: a%.    415 tone ;
: A.     440 tone ;
: a#.    466 tone ;
: b%.    466 tone ; 
: B.     494 tone ;
: C'     523 tone ;  \ octave upper 1
: c#'    554 tone ;
: d%'    554 tone ;
: D'     587 tone ;
: d#'    622 tone ;
: e%'    622 tone ;
: E'     659 tone ;
: F'     698 tone ;
: f#'    740 tone ;
: g%'    740 tone ;
: G'     784 tone ;
: g#'    831 tone ;
: a%'    831 tone ;
: A'     880 tone ;
: a#'    932 tone ;
: b%'    932 tone ;
: B'     988 tone ;
: C''   1047 tone ;  \ octave upper 2
: c#''  1109 tone ;
: d%''  1109 tone ;
: D''   1175 tone ;
: d#''  1245 tone ;
: e%''  1245 tone ;
: E''   1319 tone ;
: F''   1397 tone ;
: f#''  1480 tone ;
: g%''  1480 tone ;
: G''   1568 tone ;
: g#''  1661 tone ;
: a%''  1661 tone ;
: A''   1760 tone ;
: a#''  1865 tone ;
: b%''  1865 tone ;
: B''   1976 tone ;
: C'''  2093 tone ;  \ octave upper 3
: c#''' 2217 tone ;
: d%''' 2217 tone ;
: D'''  2349 tone ;
: d#''' 2489 tone ;
: e%''' 2489 tone ; 
: E'''  2637 tone ;
: F'''  2794 tone ;
: f#''' 2960 tone ;
: g%''' 2960 tone ;
: G'''  3126 tone ;
: g#''' 3322 tone ;
: a%''' 3322 tone ;
: A'''  3520 tone ;
: a#''' 3729 tone ;
: b%''' 3729 tone ;
: B'''  3951 tone ;  \ 12-bits of frequency goes up to B'''
: C'''' 4186 tone ;  \ octave upper 4

\  === REPLACED by full array of notes
\ * SIMPLE NUMBERS TO NOTES via an array of note executables *
\ create notes ' c. , ' d. , ' e. , ' f. , ' g. , ' a. , ' b. , ' c' , 
\ : .. ( n -- )  \ plays note from array
\  1 - dwords notes + @ execute ;


: help2 ( -- )
    cr ."  2.  13 out       - set pin 13 to out"
    cr ."  3.  2 high        - sends voltage to pin 2  (choose pins 2 through 12)"
    cr ."  4.  2 low         - sends voltage to pin 2 (choose pins 2 through 12)"
    cr ."  5.  13 setspeaker    - identifies the speaker"
    cr ."  6. 440 tone"
    cr ."  7.  1 .. c... c.. c. d. e. c' c'' f'# -  sends tone to pin 13 (speaker)  c. is middle, c.. is octave lower, c' is octave higher "
;


: help-scales
  cr ." Western styles: equitemper  cstd  cmaj"
  cr ." Chinese styles (5): yu yu-1(*) shang gong jue zhi"
  cr ." Indian styles (8): bhairav, asavari, bhairavi, kafi, khamaj, kalyan, purvi, todi"
;


: help-misc
    cr ."           440 tone      - sends 440 Hz (A) note to speaker"
    cr ."    c. d. e. c' c'' c.. c... f'# a.% -  sends tone to pin 13 (speaker)  c. is middle, c.. is octave lower, c' is octave higher, # sharp, % flat
    cr ."    1 .. 2 ..  - convert numbered simple piano to notes"
    
    cr ."  playlist   - runs through wings, harp, ode, memory, daffodils, morning, stairway, freedom, rainbow, reason, scarborough "
    cr ."  compositions: Assad: yearning snake  Jasmine: tog (together) "
    cr ."  => SOUND / MUSIC (tone)   include 3iSounds.fs"  
    cr ."       start-speaker  - sets pin 13 to speaker" 
    cr ."       record         - toggles recording mode for the keys (turns digital piano serial transmit on/off)"
    cr ."       save-bar       - saves 5 notes"
    cr ."       play-notes     - plays all saved notes"
    cr ."       delete-note    - delete last saved note"

;

: help-sound
    cr
    cr ." HELP-SOUND"
    cr ." ----------"
    cr
    cr ." Get started:"
    cr ."     plug in digital piano"
    cr ."     use RealTerm to open the connection to the port"
    cr ." In GForth:"
    cr ."     include 3iForth.fs   include 3iPort4x4.fs   include 3iSound4x4.fs"
    cr ."     5 open-com  start-piano4x4"
    cr ." Several ways to play:"
    cr ."    1.  c.  c.. c' c'' c''#  b.. b..%  a.  - play note, any octave, sharp or flat"
    cr ."    2.  440 tone  - play  frequency 440Hz A"
    cr ."    3.  45 note - play 45th note (440Hz A) from zero point, organized in octaves of 12 notes each"
    cr ."    4.  3 0 set-zero    0 note    12 note   c.   a.  9 note  "
    cr ."    5.                  1 .,  13 .,  a.  10 .,  - plays counting from 1 (lights on) "
    cr ."    reset-zero 46 ., "
    cr ."    7.  cstd  play-scale   1 .. 6 ..   - plays the notes on the scale of 8 preferred keys"
    cr ."    8.  cmaj  play-scale  - change scale"
    cr ."    9.  cstd  1 2 3  1 3 1 3  7 play-bar - plays n notes from stack"
    cr ." More help:  help-scales  help-record   help-playlist
;


\ ****************************
help-sound
